/** @file ex.h
 *
 *  Interface to GiNaC's light-weight expression handles. */

/*
 *  GiNaC Copyright (C) 1999-2001 Johannes Gutenberg University Mainz, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __GINAC_EX_H__
#define __GINAC_EX_H__

#include "basic.h"
#include "operators.h"

namespace GiNaC {

// Sorry, this is the only constant to pollute the global scope, the other ones
// are defined in utils.h and not visible from outside.
class ex;
extern const ex & _ex0(void);     ///<  single ex(numeric(0))

class symbol;
class lst;
class scalar_products;

/** Lightweight wrapper for GiNaC's symbolic objects.  Basically all it does is
 *  to hold a pointer to the other objects, manage the reference counting and
 *  provide methods for manipulation of these objects.  (Some people call such
 *  a thing a proxy class.) */
class ex
{
	friend class basic;
	
// member functions
	
	// default ctor, dtor, copy ctor assignment operator and helpers
public:
	ex();
	~ex();
	ex(const ex & other);
	const ex & operator=(const ex & other);
	// other ctors
public:
	ex(const basic & other);
	ex(int i);
	ex(unsigned int i);
	ex(long i);
	ex(unsigned long i);
	ex(double const d);
	/** Construct ex from string and a list of symbols. The input grammar is
	 *  similar to the GiNaC output format. All symbols to be used in the
	 *  expression must be specified in a lst in the second argument. Undefined
	 *  symbols and other parser errors will throw an exception. */
	ex(const std::string &s, const ex &l);
	
	// functions overriding virtual functions from bases classes
	// none
	
	// new virtual functions which can be overridden by derived classes
	// none

	// non-virtual functions in this class
public:
	void swap(ex & other);
	void print(const print_context & c, unsigned level = 0) const;
	void printtree(std::ostream & os) const;
	void dbgprint(void) const;
	void dbgprinttree(void) const;
	bool info(unsigned inf) const;
	unsigned nops() const;
	ex expand(unsigned options=0) const;
	bool has(const ex & other) const;
	int degree(const ex & s) const;
	int ldegree(const ex & s) const;
	ex coeff(const ex & s, int n=1) const;
	ex lcoeff(const ex & s) const { return coeff(s, degree(s)); }
	ex tcoeff(const ex & s) const { return coeff(s, ldegree(s)); }
	ex numer(void) const;
	ex denom(void) const;
	ex unit(const symbol &x) const;
	ex content(const symbol &x) const;
	numeric integer_content(void) const;
	ex primpart(const symbol &x) const;
	ex primpart(const symbol &x, const ex &cont) const;
	ex normal(int level = 0) const;
	ex to_rational(lst &repl_lst) const;
	ex smod(const numeric &xi) const;
	numeric max_coefficient(void) const;
	ex collect(const ex & s) const;
	ex eval(int level = 0) const;
	ex evalf(int level = 0) const;
	ex diff(const symbol & s, unsigned nth = 1) const;
	ex series(const ex & r, int order, unsigned options = 0) const;
	ex subs(const lst & ls, const lst & lr) const;
	ex subs(const ex & e) const;
	exvector get_free_indices(void) const;
	ex simplify_indexed(void) const;
	ex simplify_indexed(const scalar_products & sp) const;
	ex simplify_ncmul(const exvector & v) const;
	ex operator[](const ex & index) const;
	ex operator[](int i) const;
	ex op(int i) const;
	ex & let_op(int i);
	ex lhs(void) const;
	ex rhs(void) const;
	int compare(const ex & other) const;
	bool is_equal(const ex & other) const;
	bool is_zero(void) const { return is_equal(_ex0()); }
	
	unsigned return_type(void) const;
	unsigned return_type_tinfo(void) const;
	unsigned gethash(void) const;
	
	ex exadd(const ex & rh) const;
	ex exmul(const ex & rh) const;
private:
	void construct_from_basic(const basic & other);
	void construct_from_int(int i);
	void construct_from_uint(unsigned int i);
	void construct_from_long(long i);
	void construct_from_ulong(unsigned long i);
	void construct_from_double(double d);
	void construct_from_string_and_lst(const std::string &s, const ex &l);
	void makewriteable();

#ifdef OBSCURE_CINT_HACK
public:
	static bool last_created_or_assigned_bp_can_be_converted_to_ex(void)
	{
		if (last_created_or_assigned_bp==0) return false;
		if ((last_created_or_assigned_bp->flags &
			 status_flags::dynallocated)==0) return false;
		if ((last_created_or_assigned_bp->flags &
			 status_flags::evaluated)==0) return false;
		return true;
	}
protected:
	void update_last_created_or_assigned_bp(void)
	{
		if (last_created_or_assigned_bp!=0) {
			if (--last_created_or_assigned_bp->refcount == 0) {
				delete last_created_or_assigned_bp;
			}
		}
		last_created_or_assigned_bp = bp;
		++last_created_or_assigned_bp->refcount;
		last_created_or_assigned_exp = (long)(void *)(this);
	}
#endif // def OBSCURE_CINT_HACK

// member variables

public:
	basic *bp;      ///< pointer to basic object managed by this
#ifdef OBSCURE_CINT_HACK
	static basic * last_created_or_assigned_bp;
	static basic * dummy_bp;
	static long last_created_or_assigned_exp;
#endif // def OBSCURE_CINT_HACK
};


// performance-critical inlined method implementations

inline
ex::ex() : bp(_ex0().bp)
{
	/*debugmsg("ex default ctor",LOGLEVEL_CONSTRUCT);*/
	GINAC_ASSERT(_ex0().bp!=0);
	GINAC_ASSERT(_ex0().bp->flags & status_flags::dynallocated);
	GINAC_ASSERT(bp!=0);
	++bp->refcount;
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::~ex()
{
	/*debugmsg("ex dtor",LOGLEVEL_DESTRUCT);*/
	GINAC_ASSERT(bp!=0);
	GINAC_ASSERT(bp->flags & status_flags::dynallocated);
	if (--bp->refcount == 0)
		delete bp;
}

inline
ex::ex(const ex & other) : bp(other.bp)
{
	/*debugmsg("ex copy ctor",LOGLEVEL_CONSTRUCT);*/
	GINAC_ASSERT(bp!=0);
	GINAC_ASSERT((bp->flags) & status_flags::dynallocated);
	++bp->refcount;
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
const ex & ex::operator=(const ex & other)
{
	/*debugmsg("ex operator=",LOGLEVEL_ASSIGNMENT);*/
	GINAC_ASSERT(bp!=0);
	GINAC_ASSERT(bp->flags & status_flags::dynallocated);
	GINAC_ASSERT(other.bp!=0);
	GINAC_ASSERT(other.bp->flags & status_flags::dynallocated);
	if (--bp->refcount==0)
		delete bp;
	++other.bp->refcount;
	bp = other.bp;
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
	return *this;
}

inline
ex::ex(const basic & other)
{
	/*debugmsg("ex ctor from basic",LOGLEVEL_CONSTRUCT);*/
	construct_from_basic(other);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::ex(int i)
{
	/*debugmsg("ex ctor from int",LOGLEVEL_CONSTRUCT);*/
	construct_from_int(i);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::ex(unsigned int i)
{
	/*debugmsg("ex ctor from unsigned int",LOGLEVEL_CONSTRUCT);*/
	construct_from_uint(i);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::ex(long i)
{
	/*debugmsg("ex ctor from long",LOGLEVEL_CONSTRUCT);*/
	construct_from_long(i);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::ex(unsigned long i)
{
	/*debugmsg("ex ctor from unsigned long",LOGLEVEL_CONSTRUCT);*/
	construct_from_ulong(i);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::ex(double const d)
{
	/*debugmsg("ex ctor from double",LOGLEVEL_CONSTRUCT);*/
	construct_from_double(d);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
ex::ex(const std::string &s, const ex &l)
{
	/*debugmsg("ex ctor from string,lst",LOGLEVEL_CONSTRUCT);*/
	construct_from_string_and_lst(s, l);
#ifdef OBSCURE_CINT_HACK
	update_last_created_or_assigned_bp();
#endif // def OBSCURE_CINT_HACK
}

inline
int ex::compare(const ex & other) const
{
	GINAC_ASSERT(bp!=0);
	GINAC_ASSERT(other.bp!=0);
	if (bp==other.bp)  // trivial case: both expressions point to same basic
		return 0;
	return bp->compare(*other.bp);
}

inline
bool ex::is_equal(const ex & other) const
{
	GINAC_ASSERT(bp!=0);
	GINAC_ASSERT(other.bp!=0);
	if (bp==other.bp)  // trivial case: both expressions point to same basic
		return true;
	return bp->is_equal(*other.bp);
}


// utility functions
inline bool are_ex_trivially_equal(const ex &e1, const ex &e2)
{
	return e1.bp == e2.bp;
}

// wrapper functions around member functions
inline unsigned nops(const ex & thisex)
{ return thisex.nops(); }

inline ex expand(const ex & thisex, unsigned options = 0)
{ return thisex.expand(options); }

inline bool has(const ex & thisex, const ex & other)
{ return thisex.has(other); }

inline int degree(const ex & thisex, const ex & s)
{ return thisex.degree(s); }

inline int ldegree(const ex & thisex, const ex & s)
{ return thisex.ldegree(s); }

inline ex coeff(const ex & thisex, const ex & s, int n=1)
{ return thisex.coeff(s, n); }

inline ex numer(const ex & thisex)
{ return thisex.numer(); }

inline ex denom(const ex & thisex)
{ return thisex.denom(); }

inline ex normal(const ex & thisex, int level=0)
{ return thisex.normal(level); }

inline ex to_rational(const ex & thisex, lst & repl_lst)
{ return thisex.to_rational(repl_lst); }

inline ex collect(const ex & thisex, const ex & s)
{ return thisex.collect(s); }

inline ex eval(const ex & thisex, int level = 0)
{ return thisex.eval(level); }

inline ex evalf(const ex & thisex, int level = 0)
{ return thisex.evalf(level); }

inline ex diff(const ex & thisex, const symbol & s, unsigned nth = 1)
{ return thisex.diff(s, nth); }

inline ex series(const ex & thisex, const ex & r, int order, unsigned options = 0)
{ return thisex.series(r, order, options); }

inline ex subs(const ex & thisex, const ex & e)
{ return thisex.subs(e); }

inline ex subs(const ex & thisex, const lst & ls, const lst & lr)
{ return thisex.subs(ls, lr); }

inline ex op(const ex & thisex, int i)
{ return thisex.op(i); }

inline ex lhs(const ex & thisex)
{ return thisex.lhs(); }

inline ex rhs(const ex & thisex)
{ return thisex.rhs(); }

inline bool is_zero(const ex & thisex)
{ return thisex.is_zero(); }

inline void swap(ex & e1, ex & e2)
{ e1.swap(e2); }

} // namespace GiNaC

#endif // ndef __GINAC_EX_H__
